#include "ARGBData.h"
#ifdef USE_PNG
#include <png.h>
#include <cstdio> // fopen() and friends
#endif
#include "Server.h"
#include "Colour.h"
#include <iostream>

namespace Visibility {

ARGBData::ARGBData(CARD32 * data, unsigned width, unsigned height) {
	// copy the stuff into our vector
	for (unsigned pos = 0; pos < width*height; pos++) {
		this->data.push_back(data[pos]);
	}

	this->width = width;
	this->height = height;
}

ARGBData::ARGBData(long * data, unsigned width, unsigned height) {
	// copy the stuff into our vector
	for (unsigned pos = 0; pos < width*height; pos++) {
		this->data.push_back(CARD32(data[pos]));
	}

	this->width = width;
	this->height = height;
}

#if USE_PNG
#define PNG_SIG_BYTES 8

ARGBData::ARGBData(const std::string & file_path) {
	FILE * file = fopen(file_path.c_str(), "rb");
	width = height = 0;
	if (!file) {
		// fail
		return;
	}
	png_byte * header = new png_byte[PNG_SIG_BYTES];
	if (!fread(header, 1, PNG_SIG_BYTES, file)) {
		delete[] header;
		fclose(file);
		return;
	}
	if (png_sig_cmp(header, 0, PNG_SIG_BYTES)) {
		delete[] header;
		fclose(file);
		return;
	}

	delete[] header;

	png_structp png = png_create_read_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
	if (!png) {
		// fail
		fclose(file);
		return;
	}

	png_infop png_info = png_create_info_struct(png);
	if (!png_info) {
		// fail
		png_destroy_read_struct(&png, (png_infopp)NULL, (png_infopp)NULL);
		return;
	}
	png_infop png_end_info = png_create_info_struct(png);
	if (!png_end_info) {
		// fail
		png_destroy_read_struct(&png, &png_info, (png_infopp)NULL);
		return;
	}
	if (setjmp(png_jmpbuf(png))) {
		png_destroy_read_struct(&png, &png_info, &png_end_info);
		fclose(file);
		// fail
		return;
	}

	// oh dear, it worked
	png_init_io(png, file);
	png_set_sig_bytes(png, PNG_SIG_BYTES);

	// ARGB, LSBFirst
	png_read_png(png, png_info, PNG_TRANSFORM_SWAP_ALPHA | PNG_TRANSFORM_STRIP_16 | PNG_TRANSFORM_PACKSWAP,
		NULL); // argh, unused parameter

	png_byte bit_depth = png_get_bit_depth(png, png_info);
	if (bit_depth != 8) {
		return;
	}
	width = png_get_image_width(png, png_info);
	height = png_get_image_height(png, png_info);

	png_bytepp row_pointers = png_get_rows(png, png_info);

	// throw the data into the 'data' vector
	for (unsigned row = 0; row < height; row++) {
		for (unsigned pixel = 0; pixel < width; pixel++) {
			unsigned base = pixel * 4;
			unsigned char blue = row_pointers[row][base + 3];
			unsigned char green = row_pointers[row][base + 2];
			unsigned char red = row_pointers[row][base + 1];
			unsigned char alpha = row_pointers[row][base];
			data.push_back((alpha << 24) | (red << 16) | (green << 8) | blue);
		}
	}

	png_destroy_read_struct(&png, &png_info, &png_end_info);
	fclose(file);
}

#endif

// returns the bit at position 'pos' in 'data'. not really. it does
// something slightly different. notice the commented out parts?
// you figure it out!
static unsigned char get_bit(unsigned char * data, unsigned pos) {
	unsigned char spec = data[pos / 8];
	//printf("we want bit %u of %x\n", pos % 8, spec);
	return spec >> /*(7 - */(pos % 8)/*)*/ & 0x1;
	//return spec >> (7 - (pos % 8)) & 0x1;
}

ARGBData::ARGBData(::Pixmap pixmap, ::Pixmap mask) {
	Server * server = Server::instance();
	int x = 0, y = 0;
	unsigned pmap_width = 0, pmap_height = 0;
	Pixmap(pixmap).get_geometry(x, y, pmap_width, pmap_height);
	if (!(pmap_width && pmap_height)) throw 1; // b0rk

	if (mask) {
		unsigned mask_width = 0, mask_height = 0;
		Pixmap(mask).get_geometry(x, y, mask_width, mask_height);
		if (!(mask_width && mask_height)) mask = 0; // b0rk
		if (!((mask_width == pmap_width) && (mask_height == pmap_height))) {
			mask = 0;
		}
	}

	XImage * image = XGetImage(server->display, pixmap, 0, 0, pmap_width,
		pmap_height, 0xffffffff, ZPixmap);

	XImage * mask_image = 0;

	CARD32 * data = (CARD32*)image->data;
	unsigned char * mask_data = 0;

	if (mask) {
		mask_image = XGetImage(server->display, mask, 0, 0, pmap_width,
			pmap_height, 0xffffffff, ZPixmap);
		mask_data = (unsigned char*)mask_image->data;
	}

	unsigned length = /*image->bytes_per_line*/pmap_width * pmap_height;

	if (mask_image) {
		for (unsigned i = 0; i < length; i++) {
			CARD32 pixel = Colour(data[i], image->bits_per_pixel).pixel_value(24);
			// clear any alpha channel
			pixel &= ~(0xff << 24);
			// set the right alpha channel
			if (get_bit(mask_data, i)) {
				pixel |= (0xff << 24);
			}
			this->data.push_back(pixel);
		}
		XDestroyImage(mask_image);
	} else {
		for (unsigned i = 0; i < length; i++) {
			this->data.push_back(Colour(data[i],
				image->bits_per_pixel).pixel_value(24));
		}
	}

	XDestroyImage(image);
	width = pmap_width;
	height = pmap_height;
}

#define FRACTION        12
#define FLOOR(i)        ((i) & (~0UL << FRACTION))
#define RrDefaultAlphaOffset 24
#define RrDefaultRedOffset 16
#define RrDefaultGreenOffset 8
#define RrDefaultBlueOffset 0

void ImageCopyResampled(CARD32 *dst, CARD32 *src,
                               unsigned long dstW, unsigned long dstH,
                               unsigned long srcW, unsigned long srcH)
{
    unsigned long dstX, dstY, srcX, srcY;
    unsigned long srcX1, srcX2, srcY1, srcY2;
    unsigned long ratioX, ratioY;

    ratioX = (srcW << FRACTION) / dstW;
    ratioY = (srcH << FRACTION) / dstH;

    srcY2 = 0;
    for (dstY = 0; dstY < dstH; dstY++) {
        srcY1 = srcY2;
        srcY2 += ratioY;

        srcX2 = 0;
        for (dstX = 0; dstX < dstW; dstX++) {
            unsigned long red = 0, green = 0, blue = 0, alpha = 0;
            unsigned long portionX, portionY, portionXY, sumXY = 0;
            CARD32 pixel;

            srcX1 = srcX2;
            srcX2 += ratioX;

            for (srcY = srcY1; srcY < srcY2; srcY += (1UL << FRACTION)) {
                if (srcY == srcY1) {
                    srcY = FLOOR(srcY);
                    portionY = (1UL << FRACTION) - (srcY1 - srcY);
                    if (portionY > srcY2 - srcY1)
                        portionY = srcY2 - srcY1;
                }
                else if (srcY == FLOOR(srcY2))
                    portionY = srcY2 - srcY;
                else
                    portionY = (1UL << FRACTION);

                for (srcX = srcX1; srcX < srcX2; srcX += (1UL << FRACTION)) {
                    if (srcX == srcX1) {
                        srcX = FLOOR(srcX);
                        portionX = (1UL << FRACTION) - (srcX1 - srcX);
                        if (portionX > srcX2 - srcX1)
                            portionX = srcX2 - srcX1;
                    }
                    else if (srcX == FLOOR(srcX2))
                        portionX = srcX2 - srcX;
                    else
                        portionX = (1UL << FRACTION);

                    portionXY = (portionX * portionY) >> FRACTION;
                    sumXY += portionXY;

                    pixel = *(src + (srcY >> FRACTION) * srcW + (srcX >> FRACTION));
                    red   += ((pixel >> RrDefaultRedOffset)   & 0xFF) * portionXY;
                    green += ((pixel >> RrDefaultGreenOffset) & 0xFF) * portionXY;
                    blue  += ((pixel >> RrDefaultBlueOffset)  & 0xFF) * portionXY;
                    alpha += ((pixel >> RrDefaultAlphaOffset) & 0xFF) * portionXY;
                }
            }

            if (sumXY == 0) return; // oh crap
            red   /= sumXY;
            green /= sumXY;
            blue  /= sumXY;
            alpha /= sumXY;

            *dst++ = (red   << RrDefaultRedOffset)   |
                     (green << RrDefaultGreenOffset) |
                     (blue  << RrDefaultBlueOffset)  |
                     (alpha << RrDefaultAlphaOffset);
        }
    }
}

static void apply_alpha(CARD32 * data, unsigned length, unsigned bg_colour) {
	unsigned pos = 0;

	unsigned char bg_blue = bg_colour & 0xff;
	unsigned char bg_green = (bg_colour >> 8) & 0xff;
	unsigned char bg_red = (bg_colour >> 16) & 0xff;

	while (pos < length) {
		unsigned char alpha = (data[pos] >> 24) & 0xff;
		unsigned char red = (data[pos] >> 16) & 0xff;
		unsigned char green = (data[pos] >> 8) & 0xff;
		unsigned char blue = data[pos] & 0xff;

		red = bg_red + (red - bg_red) * alpha / 255;
		green = bg_green + (green - bg_green) * alpha / 255;
		blue = bg_blue + (blue - bg_blue) * alpha / 255;

		data[pos] = (0xff << 24) | (red << 16) | (green << 8) | blue;

		pos++;
	}
}

static void reduce_depth(CARD32 * image_data, unsigned length, unsigned char depth) {
	switch(depth) {
		case 24:
			break;
		case 16:
			CARD16 * image_data_16 = (CARD16*) image_data;
			for (unsigned i = 0; i < length; i++) {
				image_data_16[i] = Colour(image_data[i]).pixel_value(16);
			}
			break;
	}
}

XImage * ARGBData::to_ximage(unsigned width, unsigned height, CARD32 bg_colour) {
	CARD32 * image_data = (CARD32*)calloc(data.size(), sizeof(CARD32));
	std::copy(data.begin(), data.end(), image_data);

	Server * server = Server::instance();

	XImage * image = XCreateImage(server->display, server->visual, server->depth, ZPixmap, 0, NULL, width, height, 32, 0);

	if (!image) return image;

	CARD32 * scaled_image_data = image_data;

	if (!(this->width == width && this->height == height)) {
		scaled_image_data = (CARD32*)calloc(width * height, sizeof(CARD32));
		ImageCopyResampled(scaled_image_data, image_data, width, height, this->width, this->height);
		free(image_data);
	}

	apply_alpha(scaled_image_data, width*height, bg_colour);
	if (server->depth == 16) {
	 	reduce_depth(scaled_image_data, width*height, server->depth);
		image->bytes_per_line = 2 * width;
	}

	image->byte_order = LSBFirst;
	image->width = width;
	image->height = height;
	image->data = (char*)scaled_image_data;


	return image;
}

ARGBData::~ARGBData() {
}

void ARGBData::apply_tint(CARD32 tint, double amount) {
	unsigned char tint_red = (tint >> 16) & 0xff;
	unsigned char tint_green = (tint >> 8) & 0xff;
	unsigned char tint_blue = tint & 0xff;

	std::vector<CARD32>::iterator pixel;
	for (pixel = data.begin(); pixel != data.end(); pixel++) {
		unsigned char alpha = (*pixel >> 24) & 0xff;
		if (alpha == 0) continue; // don't hint blank spots
		unsigned char red = (*pixel >> 16) & 0xff;
		unsigned char green = (*pixel >> 8) & 0xff;
		unsigned char blue = *pixel & 0xff;

		red = (unsigned char)(red * (1.0 - amount) + tint_red * amount);
		green = (unsigned char)(green * (1.0 - amount) + tint_green * amount);
		blue = (unsigned char)(blue * (1.0 - amount) + tint_blue * amount);

		*pixel = (alpha << 24) | (red << 16) | (green << 8) | blue;
	}
}

}
