#include "Config.h"
#include "ConfigParser.h"
#include "Colour.h"
#include <iostream>
#include <fstream>

// for mkdir
#include <sys/stat.h>
#include <sys/types.h>

namespace Visibility {

Config * Config::pinstance = 0;
Config * Config::instance(void) {
	if (!pinstance) {
		pinstance = new Config;
	}
	return pinstance;
}

void Config::parse_file(std::string path) {
	ConfigParser parser;
   	if (!parser.parse(path)) {
		std::cout << "warning: could not open " << path << std::endl;
   	}

	if (parser.has_key("theme")) {
		// parse this file first.
		parse_file(std::string(getenv("HOME")) + "/.themes/" + parser.get_value_string("theme") + "/visibility/theme");
	}

	if (parser.has_key("orientation")) {
		orientation = parser.get_value_string("orientation");
	}

	// a map of unsigned integer config value -> its text config value
	std::map<unsigned*, std::string> integers;
	integers[&image_size] = "image_size";
	integers[&spacing] = "spacing";
	integers[&desktop_separation] = "desktop_separation";
	integers[&tooltip_padding] = "tooltip_padding";
	integers[&text_spacing] = "text_spacing";
	integers[&border_width] = "border_width";
	integers[&gap_x] = "gap_x";
	integers[&gap_y] = "gap_y";

	std::map<unsigned*, std::string>::const_iterator int_iter;
	for (int_iter = integers.begin(); int_iter != integers.end(); int_iter++) {
		if (parser.has_key(int_iter->second)) {
			*int_iter->first = parser.get_value_int(int_iter->second);
		}
	}
	integers.clear();

	// strings

	if (parser.has_key("font")) font_name = parser.get_value_string("font");
	if (parser.has_key("tooltip_font")) tooltip_font_name = parser.get_value_string("tooltip_font");

	// booleans

	std::map<bool*, std::string> booleans;
	booleans[&show_desktop_names] = "show_desktop_names";
	booleans[&single_desktop_mode] = "single_desktop_mode";
	booleans[&set_partial_strut] = "set_partial_strut";
	booleans[&set_window_type] = "set_window_type";

	std::map<bool*, std::string>::const_iterator bool_iter;
	for (bool_iter = booleans.begin(); bool_iter != booleans.end(); bool_iter++) {
		if (parser.has_key(bool_iter->second)) {
			*bool_iter->first = parser.get_value_bool(bool_iter->second);
		}
	}
	booleans.clear();

	// colours
	std::map<CARD32*, std::string> colours;
	colours[&border_colour] = "border_colour";
	colours[&bg_colour] = "bg_colour";
	colours[&active_text_colour] = "active_text_colour";
	colours[&inactive_text_colour] = "inactive_text_colour";
	colours[&active_bg_colour] = "active_bg_colour";
	colours[&inactive_bg_colour] = "inactive_bg_colour";
	colours[&tooltip_bg_colour] = "tooltip_bg_colour";
	colours[&tooltip_text_colour] = "tooltip_text_colour";
	colours[&tooltip_border_colour] = "tooltip_border_colour";
	colours[&active_tint] = "active_tint";
	colours[&inactive_tint] = "inactive_tint";
	colours[&iconified_tint] = "iconified_tint";

	std::map<CARD32*, std::string>::const_iterator colour_iter;
	for (colour_iter = colours.begin(); colour_iter != colours.end(); colour_iter++) {
		if (parser.has_key(colour_iter->second)) {
			std::string colour_str = parser.get_value_string(colour_iter->second);
			*colour_iter->first = Colour(colour_str).pixel_value();
		}
	}
	colours.clear();

	// doubles
	std::map<double*, std::string> doubles;
	doubles[&tooltip_time] = "tooltip_time";
	doubles[&active_tint_amount] = "active_tint_amount";
	doubles[&inactive_tint_amount] = "inactive_tint_amount";
	doubles[&iconified_tint_amount] = "iconified_tint_amount";

	std::map<double*, std::string>::const_iterator double_iter;
	for (double_iter = doubles.begin(); double_iter != doubles.end(); double_iter++) {
		if (parser.has_key(double_iter->second)) {
			*double_iter->first = parser.get_value_double(double_iter->second);
		}
	}
	doubles.clear();
}

void Config::font_setup(void) {
	if (font) delete font;
	if (tooltip_font) delete tooltip_font;
	font = new Font(font_name);
	tooltip_font = new Font(tooltip_font_name);
}

void Config::parse_icon_overrides(std::string path) {
	icon_overrides.clear();
	std::fstream stream(path.c_str(), std::ios_base::in);
	if (!stream.is_open() || stream.bad()) return;
	while (!stream.eof()) {
		std::string line;
		std::getline(stream, line);
		std::string::size_type pos;
		if ((pos = line.find_first_of("#")) != std::string::npos) {
			line = std::string(line, 0, pos); // comment
		}

		pos = line.find('"', 1);
		if (pos == std::string::npos) continue;
		std::string key(line, 1, pos - 1);
		pos = line.find_first_not_of(" \t", pos + 1);
		if (pos == std::string::npos) continue;
		std::string value(line, pos);
		icon_overrides[key] = value;
	}
}

std::string Config::get_icon_override(std::string& name) {
	std::map<std::string, std::string>::const_iterator iter;
	for (iter = icon_overrides.begin(); iter != icon_overrides.end(); iter++) {
		if (iter->first == name) return iter->second;
	}

	return std::string(); // not found
}

void Config::reload(void) {
	set_defaults();
	parse_file(std::string(config_file));
	parse_icon_overrides(std::string(home_path + OVERRIDES_FILE));
}

void Config::set_defaults(void) {
	// set default values
	image_size = 16;
	spacing = 1;
	border_width = 1;
	bg_colour = 0x292929;
	inactive_bg_colour = bg_colour;
	active_bg_colour = 0xd2d2d2; // m4remockup colour
	border_colour = 0x7e7e7e;
	desktop_separation = 10;
	font_name = "sans-7:bold";
	tooltip_font_name = "sans-7";
	show_desktop_names = true;
	single_desktop_mode = false;
	set_window_type = true;
	active_text_colour = 0;
	inactive_text_colour = 0xb0b0b0;
	orientation = "bottom_left";
	gap_x = gap_y = 0;
	set_partial_strut = false;
	tooltip_padding = 3;
	tooltip_bg_colour = 0xffffff;
	tooltip_border_colour = 0;
	tooltip_text_colour = 0;
	tooltip_time = 0.5;
	text_spacing = 3;

	active_tint = inactive_tint = iconified_tint = 0;
	active_tint_amount = inactive_tint_amount = iconified_tint_amount = 0;
}

Config::Config(void) : home_path(getenv("HOME")) {
	config_file = home_path + CONFIG_FILE;
	font = tooltip_font = 0;
	reload();
}

void Config::set_config_file(std::string file) {
	if (!file.length()) return;
	if (file[0] != '/') {
		file = home_path + "/.config/visibility/" + file;
	}
	config_file = file;
	reload();
}

Config::~Config() {
	if (font) delete font;
	if (tooltip_font) delete tooltip_font;
}

}
