#ifndef MAINLOOP_HEADER_INCLUDED
#define MAINLOOP_HEADER_INCLUDED

#include <sys/time.h>
#include <vector>
#include <map>

#include <tr1/memory>

namespace MainLoop {

	// wraps struct timeval :)
	class Time {
	public:
		// uses gettimeofday() to get the current time
		Time(void) {
			gettimeofday(&timeval, 0);
		}

		Time(struct timeval & timeval) {
			set(timeval);
		};

		Time(time_t seconds, suseconds_t microseconds = 0) {
			set(seconds, microseconds);
		}

		inline void set(struct timeval & timeval) {
			this->timeval = timeval;
		}

		inline void set(time_t seconds = 0, suseconds_t microseconds = 0) {
			timeval.tv_sec = seconds + (microseconds / (unsigned)1e6);
			timeval.tv_usec = microseconds % (unsigned)1e6;
		}

		inline void get(struct timeval & timeval) {
			timeval = this->timeval;
		}

		// ye olde operators. these rock.
		inline bool operator<(Time const & rhs) const {
			return timercmp(&this->timeval, &rhs.timeval, <);
		}

		inline bool operator>(Time const & rhs) const {
			return timercmp(&this->timeval, &rhs.timeval, >);
		}

		inline bool operator<=(Time const & rhs) const {
			return timercmp(&this->timeval, &rhs.timeval, <=);
		}

		inline bool operator>=(Time const & rhs) const {
			return timercmp(&this->timeval, &rhs.timeval, >=);
		}

		inline bool operator==(Time const & rhs) const {
			return timercmp(&this->timeval, &rhs.timeval, ==);
		}

		inline Time operator+(Time const & rhs) const {
			suseconds_t microseconds =
				(this->timeval.tv_usec + rhs.timeval.tv_usec);

			time_t seconds = this->timeval.tv_sec + rhs.timeval.tv_sec
				+ (microseconds / (unsigned)1e6);

			return Time(seconds, microseconds % (unsigned)1e6);
		}

		inline Time operator+=(Time const & rhs) {
			suseconds_t microseconds =
				(this->timeval.tv_usec + rhs.timeval.tv_usec);

			time_t seconds = this->timeval.tv_sec + rhs.timeval.tv_sec
				+ (microseconds / (unsigned)1e6);

			this->timeval.tv_usec = microseconds % (unsigned)1e6;
			this->timeval.tv_sec = seconds;

			return *this;
		}

		inline Time operator-(Time const & rhs) {
			suseconds_t microseconds = this->timeval.tv_usec
				- rhs.timeval.tv_usec;
			time_t seconds = this->timeval.tv_sec - rhs.timeval.tv_sec;
			while (microseconds < 0) {
				microseconds += (unsigned)1e6;
				seconds -= 1;
			}
			return Time(seconds, microseconds);
		}

	protected:
		struct timeval timeval;
	};

	// mini-callback implementation, not flexible at all :)
	class FunctorBase {
	public:
		FunctorBase() {
			disconnect_ = false;
		}

		virtual bool operator()() const = 0;
		void disconnect(void) {
			disconnect_ = true;
		}
		virtual ~FunctorBase(){};
	protected:
		bool disconnect_;
	};

	template<class ArgT> class Functor2Base {
	public:
		Functor2Base() {
			disconnect_ = false;
		}
		virtual bool operator()(ArgT) const = 0;
		virtual ~Functor2Base(){};
		void disconnect(void) {
			disconnect_ = true;
		}

	protected:
		bool disconnect_;

	};

	// used for member functions

	template<class T> class MemFunctor : public FunctorBase {
	public:
		// these have to be defined in here, otherwise the compiler cries
		MemFunctor(bool (T::*function)(), T * object) {
			function_ = function;
			object_ = object;
		}

		inline bool operator()() const {
			if (this->disconnect_) return false; // false return value means d/c
			return (object_->*function_)();
		}

	protected:
		bool (T::*function_)();
		T * object_;
	};

	template<class T, class ArgT> class MemFunctor2 : public Functor2Base<ArgT> {
	public:
		// these have to be defined in here, otherwise the compiler cries
		MemFunctor2(bool (T::*function)(ArgT), T * object) {
			function_ = function;
			object_ = object;
		}

		inline bool operator()(ArgT arg) const {
			if (this->disconnect_) return false;
			return (object_->*function_)(arg);
		}

	protected:
		bool (T::*function_)(ArgT);
		T * object_;
	};

	class PtrFunctor : public FunctorBase {
	public:
		PtrFunctor(bool (*function)());
		bool operator()(void) const;
	
	protected:
		bool (*function_)();
	};

	template<class ArgT> class PtrFunctor2 : public Functor2Base<ArgT> {
	public:
		PtrFunctor2(bool (*function)(ArgT)) {
			function_ = function;
		}
		bool operator()(ArgT arg) const {
			if (this->disconnect_) return false;
			return (*function_)(arg);
		}

	protected:
		bool (*function_)(ArgT);
	};

	// used internally by the main loop to store timeout function info.
	class Timeout {
	public:
		// interval in microseconds
		Timeout(std::tr1::shared_ptr<FunctorBase> functor, unsigned interval);
		~Timeout();

		// calls the 'slot' and updates the next call time if necessary.
		bool tick(void);

		Time next_call_time;

	protected:
		std::tr1::shared_ptr<FunctorBase> functor_;
		unsigned interval_;
	};

	// mem_fun and ptr_fun, like sigc++!

	template<class T, class ArgT> std::tr1::shared_ptr< MemFunctor2<T, ArgT> > mem_fun (
																																bool (T::*function)(ArgT), T * object) {

		return std::tr1::shared_ptr<MemFunctor2<T, ArgT> >(new MemFunctor2<T, ArgT>(function, object));
	}

	template<class T> std::tr1::shared_ptr< MemFunctor<T> > mem_fun(
																						bool (T::*function)(), T * object) {

		return std::tr1::shared_ptr< MemFunctor<T> >(new MemFunctor<T>(function, object));
	}

	std::tr1::shared_ptr<PtrFunctor> ptr_fun(bool (*function)());

	template<class ArgT> std::tr1::shared_ptr< PtrFunctor2<ArgT> > ptr_fun( bool (*function)(ArgT)) {
	
		return std::tr1::shared_ptr< PtrFunctor2<ArgT> >(new PtrFunctor2<ArgT>(function));
	}

	// the main loop. assume there is only one of these.
	class MainLoop {
	public:
		static void run(void);

		// these functions will stop being called when they return 'false'
		static inline void add_timeout_function(std::tr1::shared_ptr<FunctorBase> functor,
																						unsigned interval) {
			timeouts.push_back(std::tr1::shared_ptr<Timeout>(new Timeout(functor, interval)));
		}

		static inline void add_idle_function(std::tr1::shared_ptr<FunctorBase> functor){
			idles.push_back(functor);
		};

		// adds a function to be called when the specified file descriptor is
		// available for reading. uses select() internally. tbh.
		static inline void add_fd_watch(int fd, std::tr1::shared_ptr< Functor2Base<int> > functor) {
			file_descriptors[fd] = functor;
		}

	protected:
		static std::vector< std::tr1::shared_ptr<Timeout> > timeouts;
		static std::vector< std::tr1::shared_ptr<FunctorBase> > idles;
		static std::map< int, std::tr1::shared_ptr< Functor2Base<int> > > file_descriptors;
		static int get_fd_set(fd_set * set);
		static void call_idle_functions(void);

		// ticks the timeouts and returns the shortest amount of time before
		// they should be ticked again - this is the time that the next timeout
		// should be called.
		static void tick_timeout_functions(Time & timeout);
	};

}

#endif
